/*
 *  hnv_set_up_tetra_flow_X11.c
 *  HeteroNumberViewer
 *
 *  Created by Naoto Morikawa on 9/23/07.
 *  Copyright (C) 2007 Naoto Morikawa. All rights reserved.
 *
 *  2007-10-01 version 0.1 release
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "hnv_main.h"
#include "hnv_computation_main.h"
#include "hnv_load_file.h"
#include "hnv_tetrahedron.h"
#include "hnv_trajectory.h"
#include "hnv_search_tree.h"
#include "hnv_cone.h"
#include "hnv_t_flow_data_strct_X11.h"	// contains #include <GLUT/glut.h>

//===============================================
// Private data
//===============================================

//===============================================
// Public data
//===============================================

struct tetra_flow my_tetrahedron_flow0;
struct tetra_flow my_tetrahedron_flow1;
struct tetra_flow my_closure_tetrahedron_flow;
struct tetra_flow * pt_my_tetrahedron_flow_org = &my_tetrahedron_flow0;
struct tetra_flow * pt_my_tetrahedron_flow_new = &my_tetrahedron_flow1;
struct tetra_flow * pt_my_tetrahedron_flow = &my_tetrahedron_flow0;


//===============================================
// Private methods
//===============================================
//
// void print_tetra_peak(struct tetra_peak * peak, int index);
// void print_tetra_tile(struct tetra_tile * tile);
// void print_tetra_seq(struct tetra_seq * seq);
// void compute_outer_product(struct tetra_vertex * normal, struct tetra_vertex * vector0,
//			      struct tetra_vertex * vector1, struct tetra_vertex * vector2);
// void set_up_tetra_tile(struct tetra_tile * cur_tile, struct tetrahedron * cur_tetra);
// void set_up_tetra_seq(struct tetra_seq * cur_seq, struct trajectory * cur_traj);
// void set_up_tetra_seq_all(struct tetra_flow * t_flow);
// void set_up_peaks(struct tetra_peak * cur_peaks, int * cur_peak_ct,
//  		    struct hnv_cone * std_cone);

void print_tetra_peak(struct tetra_peak * peak, int index)
{
	fprintf(STDOUT, "print_tetra_peak() : coor4d[%d] = { %8.3f, %8.3f, %8.3f, %8.3f }, ",
		index, peak->coor4d[0], peak->coor4d[1], peak->coor4d[2], peak->coor4d[3]);
	switch (peak->redundance) {
		case HNV_REDUNDANT:
			fprintf(STDOUT, "redundance = REDUNDANT\n");
			break;
		case HNV_NOT_REDUNDANT:
			fprintf(STDOUT, "redundance = NOT_REDUNDANT\n");
			break;
		case HNV_UNKNOWN_P:
			fprintf(STDOUT, "redundance = UNKNOWN\n");
			break;
		default:
			fprintf(STDOUT, "ERROR redundance = invalid value\n");
			break;
	}
	fprintf(STDOUT, "print_tetra_peak() : coor3d[%d] = { %8.3f, %8.3f, %8.3f }, ",
		index, peak->coor3d[0], peak->coor3d[1], peak->coor3d[2]);
	return;
}


void print_tetra_tile(struct tetra_tile * tile)
{
	int i;
	
	for (i = 0; i < 4; i++) {
		fprintf(STDOUT, "print_tetra_tile() : vector[%d] = { %8.3f, %8.3f, %8.3f }, normal[%d] = { %8.3f, %8.3f, %8.3f }\n",
			i, tile->vertices[i].coor[0], tile->vertices[i].coor[1], tile->vertices[i].coor[2],
			i, tile->normals[i].coor[0], tile->normals[i].coor[1], tile->normals[i].coor[2]);
	}
	fprintf(STDOUT, "print_tetra_tile() : direction = { %8.3f, %8.3f, %8.3f }, length = %8.3f,",
		tile->direction_vector.coor[0], tile->direction_vector.coor[1], 
		tile->direction_vector.coor[2], tile->direction_vector_length);
	switch (tile->chiral) {
		case HNV_RIGHT_CHIRAL:
			fprintf(STDOUT, " chirality = RIGHT\n");
			break;
		case HNV_LEFT_CHIRAL:
			fprintf(STDOUT, " chirality = LEFT\n");
			break;
		case HNV_UNKNOWN_CHIRAL:
			fprintf(STDOUT, " chirality = UNKNOWN\n");
			break;
		default:
			fprintf(STDOUT, " ERROR chirality = invalid value\n");
			break;
	}
	switch (tile->gradient) {
		case HNV_XYZ:
			fprintf(STDOUT, "print_tetra_tile() : gradient = XYZ,");
			break;
		case HNV_XYW:
			fprintf(STDOUT, "print_tetra_tile() : gradient = XYW,");
			break;
		case HNV_XZW:
			fprintf(STDOUT, "print_tetra_tile() : gradient = XZW,");
			break;
		case HNV_YZW:
			fprintf(STDOUT, "print_tetra_tile() : gradient = YZW,");
			break;
		case HNV_UNKNOWN_G:
			fprintf(STDOUT, "print_tetra_tile() : gradient = UNKNOWN,");
			break;
		default:
			fprintf(STDOUT, "print_tetra_tile() : ERROR gradient = invalid value\n");
			break;
	}
	switch (tile->belongs_to) {
		case HNV_IN:
			fprintf(STDOUT, " belongs_to = IN,");
			break;
		case HNV_OUT:
			fprintf(STDOUT, " belongs_to = OUT,");
			break;
		case HNV_BD:
			fprintf(STDOUT, " belongs_to = BD,");
			break;
		case HNV_UNKNOWN_C:
			fprintf(STDOUT, " belongs_to = UNKNOWN,");
			break;
		default:
			fprintf(STDOUT, "\nprint_tetra_tile() : ERROR belongs_to = invalid value\n");
			break;
	}
	switch (tile->direction) {
		case HNV_UP:
			fprintf(STDOUT, " direction = UP\n");
			break;
		case HNV_DOWN:
			fprintf(STDOUT, " direction = DOWN\n");
			break;
		case 0:
			fprintf(STDOUT, " direction = UNKNOWN\n");
			break;
		default:
			fprintf(STDOUT, "\nprint_tetra_tile() : ERROR direction = invalid value\n");
			break;
	}
	fprintf(STDOUT, "print_tetra_tile() : nearby_peak = %d\n", tile->nearby_peak);
	return;
}


void print_tetra_seq(struct tetra_seq * seq)
{
	int t_ct;
	
	switch (seq->type) {
		case HNV_CLOSED:
			fprintf(STDOUT, "print_tetra_seq() : START of tetra_seq (type = CLOSED,");
			break;
		case HNV_NOT_CLOSED:
			fprintf(STDOUT, "print_tetra_seq() : START of tetra_seq (type = OPEN,");
			break;
		case HNV_UNKNOWN_T:
			fprintf(STDOUT, "print_tetra_seq() : START of tetra_seq (type = UNKNOWN,");
			break;
		default:
			fprintf(STDOUT, "print_tetra_seq() : ERROR  start of tetra_seq (type = invalid value,");
			break;
	}
	fprintf(STDOUT, " length=%d)\n", seq->length);
	for (t_ct = 0; t_ct < seq->length; t_ct++) {
		fprintf(STDOUT, "print_tetra_seq() : Start of %d-th tile\n", t_ct);
		print_tetra_tile(seq->tiles + t_ct);
	}
	return;
}


// result = (vector1 - vector0) X (vector2 - vector0)   (outer product)
// where A x B = (Ay*Bz - Az*By, Az*Bx - Ax*Bz, Ax*By - Ay*Bx)
void compute_outer_product(struct tetra_vertex * result, struct tetra_vertex * vector0,
			   struct tetra_vertex * vector1, struct tetra_vertex * vector2)
{
	struct tetra_vertex vector01;
	struct tetra_vertex vector02;
	
	vector01.coor[0] = vector1->coor[0] - vector0->coor[0];
	vector01.coor[1] = vector1->coor[1] - vector0->coor[1];
	vector01.coor[2] = vector1->coor[2] - vector0->coor[2];
	
	vector02.coor[0] = vector2->coor[0] - vector0->coor[0];
	vector02.coor[1] = vector2->coor[1] - vector0->coor[1];
	vector02.coor[2] = vector2->coor[2] - vector0->coor[2];
	
	result->coor[0] = vector01.coor[1] * vector02.coor[2] - vector01.coor[2] * vector02.coor[1];
	result->coor[1] = vector01.coor[2] * vector02.coor[0] - vector01.coor[0] * vector02.coor[2];
	result->coor[2] = vector01.coor[0] * vector02.coor[1] - vector01.coor[1] * vector02.coor[0];
	return;
}


void set_up_tetra_tile(struct tetra_tile * cur_tile, struct tetrahedron * next_tetra)
{
	{	// Set vertices and etc.
		struct hnv_coor_3d work_vertex;
		GLfloat length;
		
		get_tetra_top_vertex(&work_vertex, next_tetra);
		cur_tile->vertices[0].coor[0] = (GLfloat) work_vertex.a;
		cur_tile->vertices[0].coor[1] = (GLfloat) work_vertex.b;
		cur_tile->vertices[0].coor[2] = (GLfloat) work_vertex.c;
		
		get_tetra_second_vertex(&work_vertex, next_tetra);
		cur_tile->vertices[1].coor[0] = (GLfloat) work_vertex.a;
		cur_tile->vertices[1].coor[1] = (GLfloat) work_vertex.b;
		cur_tile->vertices[1].coor[2] = (GLfloat) work_vertex.c;
		
		get_tetra_third_vertex(&work_vertex, next_tetra);
		cur_tile->vertices[2].coor[0] = (GLfloat) work_vertex.a;
		cur_tile->vertices[2].coor[1] = (GLfloat) work_vertex.b;
		cur_tile->vertices[2].coor[2] = (GLfloat) work_vertex.c;
		
		get_tetra_bottom_vertex(&work_vertex, next_tetra);
		cur_tile->vertices[3].coor[0] = (GLfloat) work_vertex.a;
		cur_tile->vertices[3].coor[1] = (GLfloat) work_vertex.b;
		cur_tile->vertices[3].coor[2] = (GLfloat) work_vertex.c;
		
		work_vertex.a = (float) (cur_tile->vertices[3].coor[0] - cur_tile->vertices[0].coor[0]);
		work_vertex.b = (float) (cur_tile->vertices[3].coor[1] - cur_tile->vertices[0].coor[1]);
		work_vertex.c = (float) (cur_tile->vertices[3].coor[2] - cur_tile->vertices[0].coor[2]);
		length = sqrt( work_vertex.a * work_vertex.a 
			       + work_vertex.b * work_vertex.b 
			       + work_vertex.c * work_vertex.c );
		cur_tile->direction_vector_length = length;
		cur_tile->direction_vector.coor[0] = (GLfloat) work_vertex.a / length;
		cur_tile->direction_vector.coor[1] = (GLfloat) work_vertex.b / length;
		cur_tile->direction_vector.coor[2] = (GLfloat) work_vertex.c / length;
		
		cur_tile->gradient = get_tetra_gradient(next_tetra);
		cur_tile->belongs_to = get_tetra_class(next_tetra);
		cur_tile->direction = get_tetra_direction(next_tetra);
		cur_tile->nearby_peak = get_tetra_nearby_peak(next_tetra);
	}
	
	{	// Set normals for GL_TRIANGLE_STRIP
			//
			// Normals of tetrahedron v0v1v2v3:
			//	normal of facet v0v1v2 = (v0 - v3) + (v2 -v3)
			//	normal of facet v1v2v3 = (v1 - v0) + (v3 -v0)
			//	normal of facet v2v3v0 = (v2 - v1) + (v0 -v1)
			//	normal of facet v3v0v1 = (v3 - v2) + (v1 -v2)
		struct tetra_vertex work_normal;
		GLfloat length;
		int i;
		
		//	normal of vertex v0 = (v0 - v3) + (v0 -v1)
		for (i = 0; i < 3; i++)
			work_normal.coor[i] = (cur_tile->vertices[0].coor[i] - cur_tile->vertices[3].coor[i])
				+ (cur_tile->vertices[0].coor[i] - cur_tile->vertices[1].coor[i]);
		length = 0.0;
		for (i = 0; i < 3; i++)
			length += work_normal.coor[i] * work_normal.coor[i];
		length = sqrt(length);
		for (i = 0; i < 3; i++)
			cur_tile->normals[0].coor[i] = work_normal.coor[i] / length;
		
		//	normal of vertex v1 = (v1 - v2) + (v1 -v0)
		for (i = 0; i < 3; i++)
			work_normal.coor[i] = (cur_tile->vertices[1].coor[i] - cur_tile->vertices[2].coor[i])
				+ (cur_tile->vertices[1].coor[i] - cur_tile->vertices[0].coor[i]);
		length = 0.0;
		for (i = 0; i < 3; i++)
			length += work_normal.coor[i] * work_normal.coor[i];
		length = sqrt(length);
		for (i = 0; i < 3; i++)
			cur_tile->normals[1].coor[i] = work_normal.coor[i] / length;
		
		//	normal of vertex v2 = (v2 - v1) + (v2 -v3)
		for (i = 0; i < 3; i++)
			work_normal.coor[i] = (cur_tile->vertices[2].coor[i] - cur_tile->vertices[1].coor[i])
				+ (cur_tile->vertices[2].coor[i] - cur_tile->vertices[3].coor[i]);
		length = 0.0;
		for (i = 0; i < 3; i++)
			length += work_normal.coor[i] * work_normal.coor[i];
		length = sqrt(length);
		for (i = 0; i < 3; i++)
			cur_tile->normals[2].coor[i] = work_normal.coor[i] / length;
		
		//	normal of vertex v3 = (v3 - v0) + (v3 -v2)
		for (i = 0; i < 3; i++)
			work_normal.coor[i] = (cur_tile->vertices[3].coor[i] - cur_tile->vertices[0].coor[i])
				+ (cur_tile->vertices[3].coor[i] - cur_tile->vertices[2].coor[i]);
		length = 0.0;
		for (i = 0; i < 3; i++)
			length += work_normal.coor[i] * work_normal.coor[i];
		length = sqrt(length);
		for (i = 0; i < 3; i++)
			cur_tile->normals[3].coor[i] = work_normal.coor[i] / length;
	}
	
	{	// Set chirality
		struct tetra_vertex edge_v1v2;
		struct tetra_vertex normal_v0v1v3;
		float inner_product;
		int i;
		
		for (i = 0; i < 3; i++)
			edge_v1v2.coor[i] = cur_tile->vertices[2].coor[i] - cur_tile->vertices[1].coor[i];
		
		// normal_v0v1v3 = (v1 - v0) X (v3 - v0)
		compute_outer_product(&normal_v0v1v3, &(cur_tile->vertices[0]),
				      &(cur_tile->vertices[1]), &(cur_tile->vertices[3]));
		
		inner_product = 0.0;
		for (i = 0; i < 3; i++)
			inner_product += edge_v1v2.coor[i] * normal_v0v1v3.coor[i];
		
		if (inner_product > 0.0)
			cur_tile->chiral = HNV_RIGHT_CHIRAL;
		else
			cur_tile->chiral = HNV_LEFT_CHIRAL;
	}
	return;
}


void set_up_tetra_seq(struct tetra_seq * cur_seq, struct trajectory * cur_traj)
{
	struct tetra_tile * tile_array;
	struct tetrahedron * next_tetra;
	int t_ct;
	
	cur_seq->length = get_traj_length(cur_traj);
	cur_seq->type = get_traj_type(cur_traj);
	tile_array = (struct tetra_tile *) malloc(cur_seq->length * sizeof(struct tetra_tile));
	if (tile_array == NULL) {
		fprintf(STDERR, 
			"set_up_tetra_seq() : ERROR allocation failed (tile_array[%d])\n", cur_seq->length);
		exit(1);
	}
	cur_seq->tiles = tile_array;
	
	next_tetra = get_traj_head(cur_traj);
	t_ct = 0;
	while (next_tetra != NULL) {
		if (t_ct >= get_traj_length(cur_traj)) {
			fprintf(STDERR, "set_up_tetra_seq() : ERROR overflow  next_tetra = %x, t_ct (=%d) > traj->length (=%d)\n",
				(unsigned int) next_tetra, t_ct, get_traj_length(cur_traj));
			exit(1);
		}
		set_up_tetra_tile(cur_seq->tiles + t_ct, next_tetra);
		next_tetra = get_tetra_next_tetra(next_tetra);
		t_ct++;
	}
}


// trajectory: open_traj_list_head, closed_traj_list_head;
void set_up_tetra_seq_all(struct tetra_flow * t_flow)
{
	struct tetra_seq * seq_array;
	int seq_ct;
	int array_ind;
	struct trajectory * next_traj;
	
	seq_ct = get_traj_traj_ct(&open_traj_list_head) + get_traj_traj_ct(&closed_traj_list_head);
	seq_array = (struct tetra_seq *) malloc(seq_ct * sizeof(struct tetra_seq));
	if (seq_array == NULL) {
		fprintf(STDERR, 
			"set_up_tetra_seq_all() : ERROR allocation failed (seq_array[%d])\n", seq_ct);
		exit(1);
	}
	t_flow->sequences = seq_array;
	t_flow->seq_ct = seq_ct;
	
	array_ind = 0;
	next_traj = get_traj_next(&closed_traj_list_head);
	while (next_traj != NULL) {
		set_up_tetra_seq(seq_array + array_ind, next_traj);
		next_traj = get_traj_next(next_traj);
		array_ind++;
	}
	next_traj = get_traj_next(&open_traj_list_head);
	while (next_traj != NULL) {
		set_up_tetra_seq(seq_array + array_ind, next_traj);
		next_traj = get_traj_next(next_traj);
		array_ind++;
	}
	return;
}


void set_up_peaks(struct tetra_peak ** cur_peaks, int * cur_peak_ct, int * cur_peak_size,
		  struct hnv_cone * std_cone)
{
	struct tetra_peak * peak_array;
	struct hnv_coor_4d peak_coor;
	int peak_size;
	int p_ct;
	
	peak_size = get_cone_peak_size(std_cone);
	if (peak_size <= 0) {
		fprintf(STDERR, 
			"set_up_peaks() : ERROR empty cone(peak_size=%d))\n",
			peak_size);
		exit(1);
	}
	peak_array = (struct tetra_peak *) malloc(peak_size * sizeof(struct tetra_peak));
	if (peak_array == NULL) {
		fprintf(STDERR, 
			"set_up_peaks() : ERROR allocation failed (peak_array[%d])\n",
			peak_size);
		exit(1);
	}
	*cur_peaks = peak_array;
	*cur_peak_ct = get_cone_peak_ct(std_cone);
	*cur_peak_size = get_cone_peak_size(std_cone);
	// {k, l, m, n} -->==> {k-l-m+n, k+l-m-n, k-l+m-n}
	for (p_ct = 0; p_ct < *cur_peak_size; p_ct++) {
		get_peak_coor(&peak_coor, std_cone, p_ct);
		peak_array[p_ct].coor4d[0] = (GLfloat) peak_coor.k;
		peak_array[p_ct].coor4d[1] = (GLfloat) peak_coor.l;
		peak_array[p_ct].coor4d[2] = (GLfloat) peak_coor.m;
		peak_array[p_ct].coor4d[3] = (GLfloat) peak_coor.n;
		peak_array[p_ct].coor3d[0] = (GLfloat) (peak_coor.k - peak_coor.l - peak_coor.m + peak_coor.n);
		peak_array[p_ct].coor3d[1] = (GLfloat) (peak_coor.k + peak_coor.l - peak_coor.m - peak_coor.n);
		peak_array[p_ct].coor3d[2] = (GLfloat) (peak_coor.k - peak_coor.l + peak_coor.m - peak_coor.n);
		peak_array[p_ct].redundance = get_peak_redundance(std_cone, p_ct);
	}
	return;
}


//===============================================
// Public methods
//===============================================

void print_tetra_flow(struct tetra_flow * t_flow)
{
	int seq_ct;
	int peak_ct;
	
	fprintf(STDOUT, "print_tetra_flow() : START of t_flow\n");
	fprintf(STDOUT, "print_tetra_flow() :     seq_ct = %d\n", t_flow->seq_ct);
	fprintf(STDOUT, "print_tetra_flow() :     std_cone_peak_ct = %d, std_cone_peak_size = %d\n",
		t_flow->std_cone_peak_ct, t_flow->std_cone_peak_size);
	fprintf(STDOUT, "print_tetra_flow() :     std_roof_peak_ct = %d, std_roof_peak_size = %d\n",
		t_flow->std_roof_peak_ct, t_flow->std_roof_peak_size);
	fprintf(STDOUT, "print_tetra_flow() :     conj_roof_peak_ct = %d, conj_roof_peak_size = %d\n",
		t_flow->conj_roof_peak_ct, t_flow->conj_roof_peak_size);
	for (seq_ct = 0; seq_ct < t_flow->seq_ct; seq_ct++) {
		fprintf(STDOUT, "print_tetra_flow() : START of %d-th sequence\n", seq_ct);
		print_tetra_seq(t_flow->sequences + seq_ct);
		fprintf(STDOUT, "print_tetra_flow() : END of %d-th sequence\n", seq_ct);
	}
	fprintf(STDOUT, "print_tetra_flow() :\n");
	
	fprintf(STDOUT, "print_tetra_flow() : START of std_cone[peak count=%d, size=%d]\n", t_flow->std_cone_peak_ct, t_flow->std_cone_peak_size);
	for (peak_ct = 0; peak_ct < t_flow->std_cone_peak_size; peak_ct++) {
		if (t_flow->std_cone[peak_ct].redundance != HNV_REDUNDANT)
			print_tetra_peak(t_flow->std_cone + peak_ct, peak_ct);
	}
	fprintf(STDOUT, "print_tetra_flow() : END of std_cone[peak count=%d, size=%d]\n", t_flow->std_cone_peak_ct, t_flow->std_cone_peak_size);
	fprintf(STDOUT, "print_tetra_flow() :\n");
	
	fprintf(STDOUT, "print_tetra_flow() : START of std_roof[peak count=%d, size=%d]\n", t_flow->std_roof_peak_ct, t_flow->std_roof_peak_size);
	for (peak_ct = 0; peak_ct < t_flow->std_roof_peak_size; peak_ct++) {
		if (t_flow->std_roof[peak_ct].redundance != HNV_REDUNDANT)
			print_tetra_peak(t_flow->std_roof + peak_ct, peak_ct);
	}
	fprintf(STDOUT, "print_tetra_flow() : END of std_roof[peak count=%d, size=%d]\n", t_flow->std_roof_peak_ct, t_flow->std_roof_peak_size);
	fprintf(STDOUT, "print_tetra_flow() :\n");
	
	fprintf(STDOUT, "print_tetra_flow() : START of conj_roof[peak count=%d, size=%d]\n", t_flow->conj_roof_peak_ct, t_flow->conj_roof_peak_size);
	for (peak_ct = 0; peak_ct < t_flow->conj_roof_peak_size; peak_ct++) {
		if (t_flow->conj_roof[peak_ct].redundance != HNV_REDUNDANT)
			print_tetra_peak(t_flow->conj_roof + peak_ct, peak_ct);
	}
	fprintf(STDOUT, "print_tetra_flow() : END of conj_roof[peak count=%d, %d])\n", t_flow->conj_roof_peak_ct, t_flow->conj_roof_peak_size);
	fprintf(STDOUT, "print_tetra_flow() :\n");
	
	return;
}


int get_tflow_seq_ct(struct tetra_flow * t_flow)
{
	return t_flow->seq_ct;
}


int check_tflow_seq_array(struct tetra_flow * t_flow)
{
	if (t_flow->sequences == NULL)
		return FAIL;
	else
		return SUCCESS;		
}


void get_tflow_std_cone_peak_coor(GLfloat * coor, struct tetra_flow * t_flow, int p_ct)
{
	coor[0] = t_flow->std_cone[p_ct].coor3d[0];
	coor[1] = t_flow->std_cone[p_ct].coor3d[1];
	coor[2] = t_flow->std_cone[p_ct].coor3d[2];
	return;
}


int get_tflow_std_cone_peak_ct(struct tetra_flow * t_flow)
{
	return t_flow->std_cone_peak_ct;
}


int get_tflow_std_cone_peak_size(struct tetra_flow * t_flow)
{
	return t_flow->std_cone_peak_size;
}


void initialize_tetra_flow(struct tetra_flow * t_flow)
{
	t_flow->seq_ct = 0;
	t_flow->sequences = NULL;
	t_flow->std_cone_peak_ct = 0;
	t_flow->std_cone_peak_size = 0;
	t_flow->std_cone = NULL;
	t_flow->std_roof_peak_ct = 0;
	t_flow->std_roof_peak_size = 0;
	t_flow->std_roof = NULL;
	t_flow->conj_roof_peak_ct = 0;
	t_flow->conj_roof_peak_size = 0;
	t_flow->conj_roof = NULL;
	return;
}


//
// "pt_my_tetrahedron_flow->sequences[i]->tiles[j]->normal[k].coor[l]"
// "pt_my_tetrahedron_flow->sequences[i]->tiles[j]->vertex[k].coor[l]"
// tetrahedron
//     face0:  normal0
//             vertex0, vertex1, vertex2
//     face1:  normal1, 
//             vertex0, vertex1, vertex3
//
void set_up_tetra_flow(struct tetra_flow * t_flow)
{	
	set_up_tetra_seq_all(t_flow);
	if (0) {
		int seq_ct;
		
		fprintf(STDOUT, "set_up_tetra_flow() : DEBUG result of set_up_tetra_seq_all(t_flow)\n");
		fprintf(STDOUT, "set_up_tetra_flow() : DEBUG num of sequences = %d\n", t_flow->seq_ct);
		for (seq_ct = 0; seq_ct < t_flow->seq_ct; seq_ct++) {
			fprintf(STDOUT, "set_up_tetra_flow() : DEBUG START of %d-th sequence\n", seq_ct);
			print_tetra_seq(t_flow->sequences + seq_ct);
			fprintf(STDOUT, "set_up_tetra_flow() : DEBUG END of %d-th sequence\n", seq_ct);
			fprintf(STDOUT, "set_up_tetra_flow():\n");
		}
		if (t_flow->seq_ct <= 0)
			fprintf(STDOUT, "set_up_tetra_flow() : DEBUG  t_flow->seq_ct = %d\n", t_flow->seq_ct);
		fprintf(STDOUT, "\n");
	}
	
	set_up_peaks(&(t_flow->std_cone), &(t_flow->std_cone_peak_ct), &(t_flow->std_cone_peak_size),
		     my_standard_cone);
	if (0) {
		int peak_ct;
		
		fprintf(STDOUT, "set_up_tetra_flow() :  DEBUG START of std_cone[%d]\n", t_flow->std_cone_peak_size);
		for (peak_ct = 0; peak_ct < t_flow->std_cone_peak_size; peak_ct++) {
			if (t_flow->std_cone[peak_ct].redundance != HNV_REDUNDANT)
				print_tetra_peak(t_flow->std_cone + peak_ct, peak_ct);
		}
		fprintf(STDOUT, "set_up_tetra_flow() :  DEBUG END of std_cone[%d]\n", t_flow->std_cone_peak_size);
		fprintf(STDOUT, "set_up_tetra_flow():\n");
	}
	set_up_peaks(&(t_flow->std_roof), &(t_flow->std_roof_peak_ct), &(t_flow->std_roof_peak_size),
		     my_standard_roof);
	if (0) {
		int peak_ct;
		
		fprintf(STDOUT, "set_up_tetra_flow() :  START of std_roof[%d]\n", t_flow->std_roof_peak_size);
		for (peak_ct = 0; peak_ct < t_flow->std_roof_peak_size; peak_ct++) {
			if (t_flow->std_roof[peak_ct].redundance != HNV_REDUNDANT)
				print_tetra_peak(t_flow->std_roof + peak_ct, peak_ct);
		}
		fprintf(STDOUT, "set_up_tetra_flow() :  END of std_roof[%d]\n", t_flow->std_roof_peak_size);
		fprintf(STDOUT, "set_up_tetra_flow():\n");
	}
	set_up_peaks(&(t_flow->conj_roof), &(t_flow->conj_roof_peak_ct), &(t_flow->conj_roof_peak_size),
		     my_conjugate_roof);
	if (0) {
		int peak_ct;
		
		fprintf(STDOUT, "set_up_tetra_flow() :  START of conj_roof[%d]\n", t_flow->conj_roof_peak_size);
		for (peak_ct = 0; peak_ct < t_flow->conj_roof_peak_size; peak_ct++) {
			if (t_flow->conj_roof[peak_ct].redundance != HNV_REDUNDANT)
				print_tetra_peak(t_flow->conj_roof + peak_ct, peak_ct);
		}
		fprintf(STDOUT, "set_up_tetra_flow() :  END of conj_roof[%d]\n", t_flow->conj_roof_peak_size);
		fprintf(STDOUT, "set_up_tetra_flow():\n");
		fprintf(STDOUT, "\n");
	}
	if (0) {
		fprintf(STDOUT, "set_up_tetra_flow() : DEBUG  START of my_tetrahedron_flow\n");
		print_tetra_flow(t_flow);
		fprintf(STDOUT, "set_up_tetra_flow() : DEBUG  END of my_tetrahedron_flow\n");
		fprintf(STDOUT, "\n");
	}
	return;
}


void free_cones_and_tetras(void)
{
	if (my_standard_cone != NULL)
		free_hnv_cone(my_standard_cone);
	if (my_standard_roof != NULL)
		free_hnv_cone(my_standard_roof);
	if (my_conjugate_roof != NULL)
		free_hnv_cone(my_conjugate_roof);
	free_trajectories(&open_traj_list_head);
	free_trajectories(&closed_traj_list_head);
	finalize_tetra_pool();
	finalize_search_node_tree();
	return;
}


void free_tetra_flow(struct tetra_flow * t_flow)
{
	int seq_ind;
	
	if (t_flow->sequences != NULL) {
		for (seq_ind = 0; seq_ind < t_flow->seq_ct; seq_ind++) {
			if (t_flow->sequences[seq_ind].tiles != NULL)
				free(t_flow->sequences[seq_ind].tiles);
		}
		free(t_flow->sequences);
		t_flow->sequences = NULL;
	}
	if (t_flow->std_cone != NULL)
		free(t_flow->std_cone);
	if (t_flow->std_roof != NULL)
		free(t_flow->std_roof);
	if (t_flow->conj_roof != NULL)
		free(t_flow->conj_roof);
	return;
}

